/*
 * NAME:
 *	signer - simple signature generator
 */
/*
 * RCSid:
 *	$Id: signer.c,v 1.5 2013/12/14 17:35:06 sjg Exp $
 *
 *	@(#) Copyright (c) 2012 Simon J. Gerraty
 *
 *	This file is provided in the hope that it will
 *	be of use.  There is absolutely NO WARRANTY.
 *	Permission to copy, redistribute or otherwise
 *	use this file is hereby granted provided that 
 *	the above copyright notice and this notice are
 *	left intact. 
 *      
 *	Please send copies of changes and bug-fixes to:
 *	sjg@crufty.net
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#ifndef OPENSSL_NO_ENGINE
#include <openssl/engine.h>
#endif

static void *Engine = NULL;

void
initialize(void)
{
    static int once;

    if (once)
	return;
    once = 1;
    //CRYPTO_malloc_init();
    //ERR_load_crypto_strings();
    OpenSSL_add_all_algorithms();
}

int
use_engine_ui(const char *engine, UI_METHOD *uip)
{
#ifndef OPENSSL_NO_ENGINE
    ENGINE *e = NULL;

    if ((e = ENGINE_by_id(engine))) {
	if (uip)
	    ENGINE_ctrl_cmd(e, "SET_USER_INTERFACE", 0, uip, 0, 1);
	if (!ENGINE_set_default(e, ENGINE_METHOD_ALL)) {
	    ENGINE_free(e);
	    return 0;
	}
	Engine = e;
	return 1;
    }
#endif
    return 0;
}

int
use_engine(const char *engine)
{
    return use_engine_ui(engine, NULL);
}

EVP_PKEY *
load_key(const char *file)
{
    BIO *f;
    EVP_PKEY *key = NULL;

    initialize();
    f = BIO_new(BIO_s_file());
    if (BIO_read_filename(f, file) > 0) {
	key = PEM_read_bio_PrivateKey(f, NULL, NULL, NULL);
#ifdef DEBUG
	fprintf(stderr, "key(%s) = %lx\n", file, (unsigned long)key);
#endif
    }
    BIO_free(f);
    return key;
}

EVP_PKEY *
load_pubkey(const char *file)
{
    X509 *x = NULL;
    BIO *f;
    EVP_PKEY *key = NULL;

    initialize();
    f = BIO_new(BIO_s_file());
    if (BIO_read_filename(f, file) > 0) {
	x = PEM_read_bio_X509_AUX(f, NULL, NULL, NULL);
	if (x)
	    key = X509_get_pubkey(x);
    }
    BIO_free(f);
    return key;
}

size_t
sign_digest_buf(EVP_PKEY *pkey,
		const char *digest,
		unsigned char *mdata, size_t mlen,
		unsigned char *buf, size_t bufsz)
{
    EVP_MD_CTX ctx;
    const EVP_MD *md = NULL;
    EVP_PKEY_CTX *pctx = NULL;
    size_t len;

#ifdef DEBUG
	fprintf(stderr, "key= %lx\ndigest= %s\nmlen=%d, blen=%d\n",
		(unsigned long)pkey,
		digest, mlen, bufsz);
#endif
    
    initialize();
    md = EVP_get_digestbyname(digest);
    EVP_DigestInit(&ctx, md);
#ifdef DEBUG
    fprintf(stderr, "md=%lx\n", (unsigned long)md);
#endif
    if (EVP_DigestSignInit(&ctx, &pctx, md, Engine, pkey)) {
	len = bufsz;
	if (EVP_PKEY_sign(pctx, buf, &len, mdata, mlen) > 0) {
#ifdef DEBUG
	    fprintf(stderr, "rc=%d\n", len);
#endif
	    return len;
	}
    }
    return 0;
}


void
sign_digest(EVP_PKEY *pkey,
	    const char *digest,
	    unsigned char *mdata, size_t mlen,
	    unsigned char **bp, size_t *bsz)
{
    unsigned char *cp, buf[BUFSIZ];
    int c;

    c = sign_digest_buf(pkey, digest, mdata, mlen, buf, sizeof(buf));
    if (c > 0) {
	cp = malloc(c);
	memcpy(cp, buf, c);
	*bp = cp;
	*bsz = c;
    }
}


size_t
verify_digest(EVP_PKEY *pkey,
	      const char *digest,
	      unsigned char *mdata, size_t mlen,
	      unsigned char *sdata, size_t slen)
{
    EVP_MD_CTX ctx;
    const EVP_MD *md = NULL;
    EVP_PKEY_CTX *pctx = NULL;
    size_t len;
    int i = -1;

    initialize();
    md = EVP_get_digestbyname(digest);
    EVP_DigestInit(&ctx, md);

    pctx = EVP_PKEY_CTX_new(pkey, NULL);
    if (!pctx)
	goto fail;;
    if (EVP_PKEY_verify_init(pctx) <= 0)
	goto fail;
    if (EVP_PKEY_CTX_set_signature_md(pctx, ctx.digest) <= 0)
	goto fail;
    i = EVP_PKEY_verify(pctx, sdata, slen, mdata, mlen);
 fail:
    EVP_PKEY_CTX_free(pctx);
    return i;
}
	    
#ifdef MAIN
#include <fcntl.h>
#include <unistd.h>
#include <err.h>

int
main(int argc, char *argv[])
{
    EVP_PKEY *key = NULL;
    const EVP_MD *md = NULL;
    EVP_MD_CTX mctx;
    const char *hname = "sha256";
    const char *ext = ".rsig";
    unsigned char buf[BUFSIZ];
    unsigned char mdata[BUFSIZ];
    size_t mlen;
    int c;
    int fd;

    initialize();
    while ((c = getopt(argc, argv, "e:x:h:k:")) != -1) {
	switch (c) {
	case 'e':
	    use_engine_ui(optarg, NULL);
	    break;
	case 'x':
	    ext = optarg;
	    break;
	case 'h':
	    hname = optarg;
	    break;
	case 'k':
	    key = load_key(optarg);
	    if (!key)
		errx(1, "cannot load: %s", optarg);
	    break;
	}
    }

    if (!key)
	errx(1, "must specify key");
    
    md = EVP_get_digestbyname(hname);
    if (!md)
	errx(1, "unknown digest: %s", hname);

    {
	const char *sname;
	const EVP_PKEY_ASN1_METHOD *ameth;

	ameth = EVP_PKEY_get0_asn1(key);
	if (ameth) {
	    EVP_PKEY_asn1_get0_info(NULL, NULL,
				    NULL, NULL, &sname, ameth);
	    printf("%s+%s\n", sname, EVP_MD_name(md));
	}
    }
    
    for (; optind < argc; optind++) {
	if ((fd = open(argv[optind], O_RDONLY)) < 0)
	    err(1, "%s", argv[optind]);
    
	EVP_DigestInit(&mctx, md);
	while ((c = read(fd, buf, sizeof(buf))) > 0) {
	    EVP_DigestUpdate(&mctx, buf, c);
	}
	mlen = sizeof(mdata);
	EVP_DigestFinal(&mctx,mdata,&mlen);
	close(fd);
	snprintf(buf, sizeof(buf), "%s%s", argv[optind], ext);
	if ((fd = open(buf, O_WRONLY|O_CREAT, 0644)) < 0)
	    err(1, "open %s", buf);
	c = sign_digest_buf(key, hname, mdata, mlen, buf, sizeof(buf));
	if (!c)
	    errx(1, "sign failure");
	write(fd, buf, c);
	close(fd);
    }
    exit(0);
    
}
#endif
